/*
   SCI Sound Resource Profiler
   By Ravi I.

   This program takes a sound file from the command line and lists all of
   the events. The header is currently ignored because I have no idea what
   to do with it yet.
*/


#include <stdlib.h>
#include <stdio.h>
#include <io.h>
#include <iostream.h>
#include <fcntl.h>


void DisplayHelp ();
void ErrorExit (char *message);
short GetIdentifier (int fileh);
void DumpEvents (int fileh);
short ReadChar (int fileh);


int main (int argc, char **argv)
{
   int infile;

   // if the user did not pass arguments, display help
   if ( argc == 1 )
   {
      DisplayHelp ();
      return 1;
   }

   // open the file and quit on error
   infile = open (argv[1], O_RDONLY | O_BINARY);
   if ( infile == -1 )
      ErrorExit ("Error: Unable to open file for input\n");

   // check for a valid sound resource
   if ( GetIdentifier(infile) != 0x0084 )
      ErrorExit ("Error: Unrecognized file format\n");

   // display the sound events
   DumpEvents (infile);

   // close the file and exit
   close (infile);
   return 0;
}


//-------------------------------------------------------------------------
//   DisplayHelp
//   Called when the program is run without command-line arguments
//   Displays a help screen with program usage

void DisplayHelp ()
{
   cout << "\nSCI Sound Resource Profiler\n";
   cout << "verison 0.01\n";
   cout << "By Ravi I.\n\n";
   cout << "Usage:\n";
   cout << "   sndprf <infile>\n";
}


//-------------------------------------------------------------------------
//   ErrorExit
//   Called when the program encounters an error and must quit
//   Displays an error message then exits

void ErrorExit (char *message)
{
   cout << message;
   exit (1);
}


//-------------------------------------------------------------------------
//   GetIdentifier
//   Called with the input file handle to check for a sound resource
//   Returns the first two bytes of the file as a short int

short GetIdentifier (int fileh)
{
   short id;

   lseek (fileh, 0, SEEK_SET);
   read (fileh, &id, 2);
   return id;
}


//-------------------------------------------------------------------------
//   DumpEvents
//   Called with the input file handle to list sound events
//   Displays one event per line

void DumpEvents (int fileh)
{
   int bytesread;
   unsigned char foo, statusbyte, p1, p2, channel;
   short bar, eventtime;
   char command[80];

   // skip the file header
   if ( lseek(fileh, 0x23, SEEK_SET) == -1 )
      ErrorExit ("Error: Unable to seek to sound events\n");

   // loop until the end of the file
   while ( (eventtime = ReadChar(fileh)) != -1 )  // get the delta time
   {

      // get the next byte
      if ( (bar = ReadChar(fileh)) == -1 )
         ErrorExit ("Error: Unexpected end of file\n");
      foo = (char)bar;

      // check for "running status byte" by checking bit 7
      if ( foo & 0x80 )
         statusbyte = foo;              // this is a command
      else
         lseek (fileh, -1, SEEK_CUR);   // rewind to read it as a paramater

      // interpret the command
      switch ( statusbyte & 0xF0 )
      {
         case 0x80:
            channel = statusbyte & 0x0F;
            if ( (bar = ReadChar (fileh)) == -1)
               ErrorExit ("Error: Unexpected end of file\n");
            p1 = (char)bar;
            if ( (bar = ReadChar (fileh)) == -1)
               ErrorExit ("Error: Unexpected end of file\n");
            p2 = (char)bar;
            sprintf (command, "8x: Note Off    Stop note %u with %u velocity on channel %u", p1, p2, channel);
            break;

         case 0x90:
            channel = statusbyte & 0x0F;
            if ( (bar = ReadChar (fileh)) == -1)
               ErrorExit ("Error: Unexpected end of file\n");
            p1 = (char)bar;
            if ( (bar = ReadChar (fileh)) == -1)
               ErrorExit ("Error: Unexpected end of file\n");
            p2 = (char)bar;
            if ( p2 > 0 )
               sprintf (command, "9x: Note On     Play note %u with velocity %u on channel %u", p1, p2, channel);
            else
               sprintf (command, "9x: Note On     Stop note (0 velocity) %u on channel %u", p1, channel);
            break;

         case 0xA0:
            channel = statusbyte & 0x0F;
            if ( (bar = ReadChar (fileh)) == -1)
               ErrorExit ("Error: Unexpected end of file\n");
            p1 = (char)bar;
            if ( (bar = ReadChar (fileh)) == -1)
               ErrorExit ("Error: Unexpected end of file\n");
            p2 = (char)bar;
            sprintf (command, "Ax: Pressure    After-touch key pressure for pitch %u to %u on channel %u", p1, p2, channel);
            break;

         case 0xB0:
            channel = statusbyte & 0x0F;
            if ( (bar = ReadChar (fileh)) == -1)
               ErrorExit ("Error: Unexpected end of file\n");
            p1 = (char)bar;
            if ( (bar = ReadChar (fileh)) == -1)
               ErrorExit ("Error: Unexpected end of file\n");
            p2 = (char)bar;
            sprintf (command, "Bx: Control     Set control %u to %u on channel %u", p1, p2, channel);
            break;

         case 0xC0:
            channel = statusbyte & 0x0F;
            if ( (bar = ReadChar (fileh)) == -1)
               ErrorExit ("Error: Unexpected end of file\n");
            p1 = (char)bar;
            sprintf (command, "Cx: Program     Set program to %u on channel %u", p1, channel);
            break;

         case 0xD0:
            channel = statusbyte & 0x0F;
            if ( (bar = ReadChar (fileh)) == -1)
               ErrorExit ("Error: Unexpected end of file\n");
            p1 = (char)bar;
            sprintf (command, "Dx: Pressure    After-touch channel pressure to %u on channel %u", p1, channel);
            break;

         case 0xE0:
            channel = statusbyte & 0x0F;
            if ( (bar = ReadChar (fileh)) == -1)
               ErrorExit ("Error: Unexpected end of file\n");
            p1 = (char)bar;
            if ( (bar = ReadChar (fileh)) == -1)
               ErrorExit ("Error: Unexpected end of file\n");
            p2 = (char)bar;
            sprintf (command, "Ex: Pitch       Set pitch wheel to %u on channel %u", p1 | (p2 << 7), channel);
            break;

         case 0xF0:
            if ( statusbyte == 0xFC )
               sprintf (command, "FC: End Sequence");
            else
            {
               sprintf (command, "Error: Unrecognized command %X\n", statusbyte);
               ErrorExit (command);
            }
            break;

         default:
            sprintf (command, "Error: Invalid command %X\n", statusbyte);
            ErrorExit (command);
      }

      cout << "Delta time: " << eventtime << "    " << command << "\n";
   }

   if ( statusbyte != 0xFC )
      ErrorExit ("Error: Unexpected end of file\n");
}


//-------------------------------------------------------------------------
//   ReadChar
//   Called with the input file handle to read a single byte
//   Returns the value or -1 on error

short ReadChar (int fileh)
{
   unsigned char charbuf;
   int bytesread;

   bytesread = read (fileh, &charbuf, 1);
   if ( bytesread != 1 )
      return -1;

   return (short)charbuf;
}