#include <stdlib.h>
#include <io.h>
#include <fcntl.h>
#include <iostream.h>
#include <sys/stat.h>


void DisplayHelp ();
void ErrorExit (char *message);
void WriteIdentifier (int fileh);
short ReadChar (int fileh);
long ReadVL (int fileh);
char GetParam (int fileh);
void CopyEvents (int inf, int outf, unsigned short tempoconst, char prog, char fixcontrols);
unsigned short ReadMThd (int filh);
void ReadMTrk (int filh);
void WriteHeader (int fileh);
void InitControls (int fileh);


void main (int argc, char **argv)
{
   unsigned short tempo;
   int infile, outfile;
   char foo, instrument = -1, fixcontrols = 0, init = 1;

   // check for command-line arguments
   if ( argc < 3 )
   {
      DisplayHelp ();
      return;
   }

   // interpret flags
   for ( int lcv=3; lcv<argc; lcv++ )
   {
      if ( argv[lcv][0] == '-' )
         switch ( argv[lcv][1] )
         {
            case 'I':
               instrument = atoi (argv[lcv+1]);
               break;

            case 'C':
               fixcontrols = 1;
               break;

            case 'M':
               init = 0;
               break;

            default:
               cout << "Unknown flag " << argv[lcv] << "\n";
         }
   }

   // open the input and output files
   _fmode = O_BINARY;
   if ( (infile = open(argv[1], O_RDONLY)) == -1 )
      ErrorExit ("Error: Unable to open MIDI file for input\n");
   if ( (outfile = creat(argv[2], S_IWRITE)) == -1 )
      ErrorExit ("Error: Unable to open sound file for output\n");

   // read the MIDI header and track header
   tempo = ReadMThd (infile);
   ReadMTrk (infile);

   // Initialize the sound resource
   WriteIdentifier (outfile);

   // initialize channel info
   WriteHeader (outfile);

   // initialize controls
   if ( init )
      InitControls (outfile);

   // copy events
   CopyEvents (infile, outfile, tempo, instrument, fixcontrols);

   // close the open files
   close (infile);
   close (outfile);
}


//-------------------------------------------------------------------------
// DisplayHelp
// Called when the program is run without argumenmts
// Displays help information

void DisplayHelp ()
{
   cout << "\nMIDI to SCI Sound Converter\n";
   cout << "version 0.02\n";
   cout << "By Ravi I.\n\n";
   cout << "Usage:\n";
   cout << "   midi2snd <midifile> <soundfile> [flags]\n\n";
   cout << "<soundfile> is the filename of an extracted sound resource\n";
   cout << "<midifile> is the output file\n\n";
   cout << "Flags must come after the filenames if included. A flag\n";
   cout << "must be separated from its paramater with a space.\n";
   cout << "   -I n     - set program to n\n";
   cout << "   -C       - remove references to unrecognized controls\n";
   cout << "   -M       - leave channels in monophonic mode\n";
}


//-------------------------------------------------------------------------
//   ErrorExit
//   Called when the program encounters an error and must quit
//   Displays an error message then exits

void ErrorExit (char *message)
{
   cout << message;
   exit (1);
}


//-------------------------------------------------------------------------
//   WriteIdentifier
//   Called with the output file handle to check for a sound resource
//   Writes the first two bytes of the file as a short int

void WriteIdentifier (int fileh)
{
   short id = 0x0084;
   write (fileh, &id, 2);
}


//-------------------------------------------------------------------------
//   ReadChar
//   Called with the input file handle to read a single byte
//   Returns the value or -1 on error

short ReadChar (int fileh)
{
   unsigned char charbuf;
   int bytesread;

   bytesread = read (fileh, &charbuf, 1);
   if ( bytesread != 1 )
      return -1;

   return (short)charbuf;
}


//-------------------------------------------------------------------------
//   ReadVL
//   Called with the output file handle and a data byte
//   Reads the variable-legth value

long ReadVL (int fileh)
{
   unsigned char foo;
   unsigned long bar = 0;
   short baz;

   while ( (baz = ReadChar(fileh)) != -1 )
   {
      foo = (char)baz;
      bar = (bar << 7) | (foo & 0x7F);
      if ( !(foo & 0x80) )
         return bar;
   }

   ErrorExit ("Error: Unexpected end of file");
   return 0xFFFF;
}


//-------------------------------------------------------------------------
//   GetParam
//   Called with the input handles
//   Gets a command paramater

char GetParam (int fileh)
{
   short foo;
   if ( (foo = ReadChar (fileh)) == -1)
      ErrorExit ("Error: Unexpected end of file\n");
   if ( foo & 0x80 )
      cout << ("Warning: Paramater with bit 7 set\n");
   return (char)foo;
}



//-------------------------------------------------------------------------
//   CopyEvents
//   Called with the input and output file handles plus special modifiers
//   Copies the sound events

void CopyEvents (int inf, int outf, unsigned short tempoconst, char prog, char fixcontrols)
{
   long deltat;
   unsigned char eventtime, statusbyte, laststatus = 0, p1, p2;
   short bar;
   unsigned long qt, micro, foo, time = 0, prevtime = 0;

   // loop until the end of the file
   while ( (deltat = ReadVL(inf)) != -1 )  // get the delta time
   {
      // scale delta t to microseconds
      micro = (deltat * qt) / tempoconst;

      // get the theoretical time for this event
      time += micro;

      // recalculate delta time based on actual time
      micro = time - prevtime;

      // convert from microseconds to sound ticks
      foo = micro / 16000;
      if ( foo > 255 )
      {
         cout << "Warning: Delta time too large - truncating\n";
         foo = 255;
      }
      eventtime = foo;

      // update prevtime with the actual elapsed time in microseconds
      micro = foo * 16000;
      prevtime += micro;

      // get the next byte
      if ( (bar = ReadChar(inf)) == -1 )
         ErrorExit ("Error: Unexpected end of file\n");
      statusbyte = (char)bar;
      p2++;

      // check for running status byte by checking bit 7
      if ( !(statusbyte & 0x80) )
      {
         statusbyte = laststatus;
         lseek (inf, -1, SEEK_CUR);   // rewind to read it as a paramater
      }

      // copy the command
      switch ( statusbyte & 0xF0 )
      {
         // these 2-paramater commands can be directly copied
         case 0x80:    // note off
         case 0x90:    // note on
         case 0xA0:    // key pressure
         case 0xE0:    // pitch wheel
            p1 = GetParam (inf);
            p2 = GetParam (inf);
            write (outf, &eventtime, 1);
            if ( statusbyte != laststatus )
               write (outf, &statusbyte, 1);
            write (outf, &p1, 1);
            write (outf, &p2, 1);
            break;

         // these 1-paramater commands can be directly copied
         case 0xD0:    // channel pressure
            p1 = GetParam (inf);
            write (outf, &eventtime, 1);
            if ( statusbyte != laststatus )
               write (outf, &statusbyte, 1);
            write (outf, &p1, 1);
            break;

         // the rest are special case commands

         // check for set MIDI instrument
         case 0xC0:    // set program
            p1 = GetParam (inf);
            if ( prog != -1 )
               p1 = (char)prog;
            write (outf, &eventtime, 1);
            if ( statusbyte != laststatus )
               write (outf, &statusbyte, 1);
            write (outf, &p1, 1);
            break;

         // check for fix controls
         case 0xB0:    // set control
            p1 = GetParam (inf);
            p2 = GetParam (inf);
            switch ( p1 )
            {
               // the recognized controls
               case 0x07:
               case 0x0A:
               case 0x40:
               case 0x4B:
               case 0x4E:
                  write (outf, &eventtime, 1);
                  if ( statusbyte != laststatus )
                     write (outf, &statusbyte, 1);
                  write (outf, &p1, 1);
                  write (outf, &p2, 1);
                  break;

               default:
                  if ( !fixcontrols )
                  {
                     write (outf, &eventtime, 1);
                     if ( statusbyte != laststatus )
                        write (outf, &statusbyte, 1);
                     write (outf, &p1, 1);
                     write (outf, &p2, 1);
                  }
            }
            break;

         case 0xF0:
            switch ( statusbyte & 0x0F )
            {
               // various events which take no data
               case 0x08:
               case 0x0A:
               case 0x0C:
                  break;

               // system exclusive messages
               case 0x00:
               case 0x07:
                  p1 = GetParam (inf);
                  lseek (inf, p1, SEEK_CUR);
                  break;

               // meta event
               case 0x0F:
                  p1 = GetParam (inf);
                  switch ( p1 )
                  {
                     // end of track
                     case 0x2F:
                        p1 = 0xFC;
                        write (outf, &eventtime, 1);
                        write (outf, &p1, 1);
                        return;

                     // microseconds per quarter note
                     case 0x51:
                        p1 = GetParam (inf);
                        if ((bar=ReadChar(inf)) == -1)
                           ErrorExit ("Error: Unexpected end of file\n");
                        qt = bar;
                        if ((bar=ReadChar(inf)) == -1)
                           ErrorExit ("Error: Unexpected end of file\n");
                        qt <<= 8;
                        qt |= bar;
                        if ((bar=ReadChar(inf)) == -1)
                           ErrorExit ("Error: Unexpected end of file\n");
                        qt <<= 8;
                        qt |= bar;
                        break;

                     default:
                        p1 = GetParam (inf);
                        lseek (inf, p1, SEEK_CUR);
                        break;
                  }
                  break;

               default:
                  ErrorExit ("Error: Unrecognized Fx command\n");
               }
            break;

         default:
            p2++;
            ErrorExit ("Error: Invalid command\n");
      }

      laststatus = statusbyte;
   }

   cout << "Warning: Unexpected end of file\n";
}



//-------------------------------------------------------------------------
//   ReadMThd
//   Called with the handle a MIDI file
//   Reads the header, returning the song tempo

unsigned short ReadMThd (int filh)
{
   short foo;
   unsigned short tempo;
   char header[] = "MThd\x00\x00\x00\x06\x00\x00\x00\x01";
   char header2[13];

   read (filh, header2, 12);
   if ( memcmp(header, header2, 12) )
      ErrorExit ("Error: Invalid file format");

   if ( (foo = ReadChar(filh)) == -1 )
      ErrorExit ("Error: Unexpected end of file");
   tempo = foo << 8;
   if ( (foo = ReadChar(filh)) == -1 )
      ErrorExit ("Error: Unexpected end of file");
   tempo |= foo;

   return tempo;
}



//-------------------------------------------------------------------------
//   ReadMTrk
//   Called with the handle to a new MIDI file
//   Reads the track header

void ReadMTrk (int outf)
{
   // I could have just incremented the file pointer, but I may eventually
   //    want to do something with this value
   char header[9];
   read (outf, header, 8);
}



//-------------------------------------------------------------------------
//   WriteHeader
//   Called with the handle a new sound file
//   Writes the sound header, initializing channels

void WriteHeader (int fileh)
{
   char foo;

   foo = 0;
   write (fileh, &foo, 1);

   for ( int lcv=0; lcv<16; lcv++)
   {
      foo = 0x01;
      write (fileh, &foo, 1);
      foo = 0x04;
      write (fileh, &foo, 1);
   }
}



//-------------------------------------------------------------------------
//   InitControls
//   Called with the handle a new sound file
//   Initializes channels

void InitControls (int fileh)
{
   char time = 0, status, p1, p2;

   for ( char lcv=0; lcv<16; lcv++ )
   {
      status = 0xB0 | lcv;
      p1 = 0x4E;
      p2 = 8;
      write (fileh, &time, 1);
      write (fileh, &status, 1);
      write (fileh, &p1, 1);
      write (fileh, &p2, 1);
   }
}
