#include <stdlib.h>
#include <io.h>
#include <fcntl.h>
#include <iostream.h>
#include <sys/stat.h>


void DisplayHelp ();
void ErrorExit (char *message);
short GetIdentifier (int fileh);
short ReadChar (int fileh);
char WriteVL (int fileh, char data);
char GetParam (int fileh);
unsigned long CopyEvents (int inf, int outf, int MIDIProg);
void WriteMThd (int outf);
void WriteMTrk (int outf);


void main (int argc, char **argv)
{
   unsigned long tracklen;
   int infile, outfile, instrument = -1;
   char foo;

   // check for command-line arguments
   if ( argc < 3 )
   {
      DisplayHelp ();
      return;
   }

   // open the input and output files
   _fmode = O_BINARY;
   if ( (infile = open(argv[1], O_RDONLY)) == -1 )
      ErrorExit ("Error: Unable to open sound file for input\n");
   if ( (outfile = creat(argv[2], S_IWRITE)) == -1 )
      ErrorExit ("Error: Unable to open midi file for output\n");

   // interpret flags
   for ( int lcv=3; lcv<argc; lcv++ )
   {
      if ( argv[lcv][0] == '-' )
         switch ( argv[lcv][1] )
         {
            case 'I':
               instrument = atoi (argv[lcv+1]);
               break;

            default:
               cout << "Unknown flag " << argv[lcv] << "\n";
         }
   }

   // check for a valid sound resource
   if ( GetIdentifier(infile) != 0x0084 )
      ErrorExit ("Error: Unrecognized file format\n");

   // write the MIDI header and track header
   WriteMThd (outfile);
   WriteMTrk (outfile);

   // copy events
   tracklen = CopyEvents (infile, outfile, instrument);

   // set the track length (written MSB first)
   lseek (outfile, 0x12, SEEK_SET);
   foo = (char)(tracklen >> 24);
   write (outfile, &foo, 1);
   foo = (char)(tracklen >> 16);
   write (outfile, &foo, 1);
   foo = (char)(tracklen >> 8);
   write (outfile, &foo, 1);
   foo = (char)(tracklen);
   write (outfile, &foo, 1);

   // close the open files
   close (infile);
   close (outfile);
}


//-------------------------------------------------------------------------
// DisplayHelp
// Called when the program is run without argumenmts
// Displays help information

void DisplayHelp ()
{
   cout << "\nSCI Sound to MIDI Converter\n";
   cout << "version 0.04\n";
   cout << "By Ravi I.\n\n";
   cout << "Usage:\n";
   cout << "   snd2midi <soundfile> <midifile> [flags]\n\n";
   cout << "<soundfile> is the filename of an extracted sound resource\n";
   cout << "<midifile> is the output file\n\n";
   cout << "Flags must come after the filenames if included. A flag\n";
   cout << "must be separated from its paramater with a space.\n";
   cout << "   -I n     - set MIDI program to n\n";
}


//-------------------------------------------------------------------------
//   ErrorExit
//   Called when the program encounters an error and must quit
//   Displays an error message then exits

void ErrorExit (char *message)
{
   cout << message;
   exit (1);
}


//-------------------------------------------------------------------------
//   GetIdentifier
//   Called with the input file handle to check for a sound resource
//   Returns the first two bytes of the file as a short int

short GetIdentifier (int fileh)
{
   short id;

   lseek (fileh, 0, SEEK_SET);
   read (fileh, &id, 2);
   return id;
}


//-------------------------------------------------------------------------
//   ReadChar
//   Called with the input file handle to read a single byte
//   Returns the value or -1 on error

short ReadChar (int fileh)
{
   unsigned char charbuf;
   int bytesread;

   bytesread = read (fileh, &charbuf, 1);
   if ( bytesread != 1 )
      return -1;

   return (short)charbuf;
}


//-------------------------------------------------------------------------
//   WriteVL
//   Called with the output file handle and a data byte
//   Writes the variable-legth value and returns no. of bytes written

char WriteVL (int fileh, char data)
{
   char num = 0, topbyte = 0x81;
   if ( data & 0x80 )
   {
      data &= 0x7F;
      write (fileh, &topbyte, 1);
      num = 1;
   }
   write (fileh, &data, 1);
   return ++num;
}


//-------------------------------------------------------------------------
//   GetParam
//   Called with the input handles
//   Gets a command paramater

char GetParam (int fileh)
{
   short foo;
   if ( (foo = ReadChar (fileh)) == -1)
      ErrorExit ("Error: Unexpected end of file\n");
   if ( foo & 0x80 )
      cout << ("Warning: Paramater with bit 7 set\h");
   return (char)foo;
}



//-------------------------------------------------------------------------
//   CopyEvents
//   Called with the input and output file handles plus special modifiers
//   Copies the sound events returning the number of bytes written

unsigned long CopyEvents (int inf, int outf, int MIDIProg)
{
   unsigned long byteswritten = 0;
   unsigned char statusbyte, laststatus = 0, p1, p2;
   short bar, eventtime;

   // skip the file header
   if ( lseek(inf, 0x23, SEEK_SET) == -1 )
      ErrorExit ("Error: Unable to seek to sound events\n");

   // loop until the end of the file
   while ( (eventtime = ReadChar(inf)) != -1 )  // get the delta time
   {
      // get the next byte
      if ( (bar = ReadChar(inf)) == -1 )
         ErrorExit ("Error: Unexpected end of file\n");
      statusbyte = (char)bar;

      // check for running status byte by checking bit 7
      if ( !(statusbyte & 0x80) )
      {
         statusbyte = laststatus;
         lseek (inf, -1, SEEK_CUR);   // rewind to read it as a paramater
      }

      // copy the command
      switch ( statusbyte & 0xF0 )
      {
         // these 2-paramater commands can be directly copied
         case 0x80:    // note off
         case 0x90:    // note on
         case 0xA0:    // key pressure
         case 0xB0:    // control change
         case 0xE0:    // pitch wheel
            p1 = GetParam (inf);
            p2 = GetParam (inf);
            byteswritten += WriteVL (outf, (char)eventtime);
            if ( statusbyte != laststatus )
            {
               write (outf, &statusbyte, 1);
               byteswritten++;
            }
            write (outf, &p1, 1);
            write (outf, &p2, 1);
            byteswritten += 2;
            break;

         // these 1-paramater commands can be directly copied
         case 0xD0:    // channel pressure
            p1 = GetParam (inf);
            byteswritten += WriteVL (outf, (char)eventtime);
            if ( statusbyte != laststatus )
            {
               write (outf, &statusbyte, 1);
               byteswritten++;
            }
            write (outf, &p1, 1);
            byteswritten += 1;
            break;

         // the rest are special case commands

         // check for set MIDI instrument
         case 0xC0:
            p1 = GetParam (inf);
            if ( MIDIProg != -1 )
               p1 = (char)MIDIProg;
            byteswritten += WriteVL (outf, (char)eventtime);
            if ( statusbyte != laststatus )
            {
               write (outf, &statusbyte, 1);
               byteswritten++;
            }
            write (outf, &p1, 1);
            byteswritten += 1;
            break;

         // the end track command
         case 0xF0:
            if ( statusbyte != 0xFC )
               ErrorExit ("Error: Unrecognized Fx command\n");
            else
               {
                  byteswritten += WriteVL (outf, (char)eventtime);
                  statusbyte = 0xFF;
                  p1 = 0x2F;
                  p2 = 0x00;
                  write (outf, &statusbyte, 1);
                  write (outf, &p1, 1);
                  write (outf, &p2, 1);
                  byteswritten += 3;
                  return byteswritten;
               }
            break;

         default:
            ErrorExit ("Error: Invalid command\n");
      }

      laststatus = statusbyte;
   }

   cout << "Warning: Unexpected end of file\n";

   return byteswritten;
}



//-------------------------------------------------------------------------
//   WriteMThd
//   Called with the handle a new MIDI file
//   Writes the header

void WriteMThd (int outf)
{
   char header[] = "MThd\x00\x00\x00\x06\x00\x00\x00\x01\x00\x20";
   write (outf, header, 14);
}



//-------------------------------------------------------------------------
//   WriteMTrk
//   Called with the handle a new MIDI file
//   Writes the track header

void WriteMTrk (int outf)
{
   char header[] = "MTrk\x00\x00\x00\x00";
   write (outf, header, 8);
}
